interface Dictionary<T> {
	//@ts-ignore
	[key: string | Symbol]: T;
}

type Constructor<T> = {
	new(...args: any[]): T;
};

type ObjectWithName = {
	name: string
}

type ProvidesConfig = { singleton?: boolean }

type Name = Symbol | string;
type NameLike<T = any> = ObjectWithName | Constructor<T> | Name;

function getName(type: NameLike): Name {
	const name = typeof type == 'function' ? type.name :
		typeof type == 'object' ? (type as ObjectWithName).name :
			type;
	return name;
}

export function adapter<T>(type: NameLike, config?: ProvidesConfig) {
	return function (target: any) {

		const name = getName(type);

		if (config && config.singleton) {
			target.singleton = name;
		} else {
			target.provides = name;
		}
	}
}


export function port<T>(type: NameLike) {
	const typeName = getName(type);

	return function (target: any, propName: string): void {
		const backendName = Symbol.for(propName);
		Reflect.defineProperty(target, backendName, {
			writable: true,
			enumerable: false
		})
		target.constructor.injection = target.constructor.injection || [];
		target.constructor.injection.push({ propName, type: typeName });
		Reflect.defineProperty(target, propName, {
			get: function () {
				if (!this[backendName]) {
					this[backendName] = injectProp(typeName, []);
				}
				return this[backendName];
			},
			set: function (value) {
				if (!this[backendName]) {
					this[backendName] = value;
				}
			}
		})
	}
}


type Resolver = () => any;

let registry: Dictionary<{ resolve: Resolver | null, instance: Constructor<any> | Object | null }> = {};

function injectProp(name: Name, args: any[]): any {
	// hack on typescript	
	if (registry[name as string]) {
		const entry = registry[name as string];
		if (entry.resolve) {
			return entry.resolve();
		} else if (entry.instance) {
			if (typeof entry.instance == 'function') {
				const ctr = entry.instance as Constructor<any>;
				entry.instance = new ctr(...args);
			}
			return entry.instance as Object;
		}
	}
}

export function New<C extends Constructor<any>>(type: C, ...args: ConstructorParameters<C>): InstanceType<C> {
	return Container.resolve(type, ...args) ?? new type(...args);
}

export class Container {

	public static clear() {
		registry = {};
	}

	public static resolve<T>(type: Constructor<T>, ...args: any[]): T;
	public static resolve<T = never>(type: Name, ...args: any[]): T extends never ? never : T;
	public static resolve<T>(type: NameLike<T>, ...args: any[]): T {
		return injectProp(getName(type), args);
	}

	public static registerAs<T>(type: Constructor<T>, name: Name): void
	public static registerAs<T extends Z, Z>(type: Constructor<T>, inter: Constructor<Z>): void
	public static registerAs<T>(type: Constructor<T>, nameOrType: NameLike): void {
		const name = getName(nameOrType) as string; // hack on symbols in dictionaries
		registry[name] = { resolve: () => new type(), instance: null };
	}

	public static registerSingleton(instance: any, name: Name)
	public static registerSingleton<T extends Z, Z>(instance: T, type: Constructor<Z>)
	public static registerSingleton(instance: any, nameOrType: NameLike) {
		const name = getName(nameOrType) as string; // hack on symbols in dictionaries
		registry[name] = { resolve: null, instance };
	}

	public static singleton<T, Z extends T = any>(type: Constructor<T>, impl?: Constructor<Z>) {
		const name = getName(type) as string; // hack on symbols in dictionaries
		registry[name] = { instance: impl ?? type };
	}

	public static register(...types: Constructor<any>[]) {
		for (const type of types) {
			let unknow = type as any;
			if (unknow.provides) {
				registry[unknow.provides] = { resolve: () => new type(), instance: null };
			} else if (unknow.singleton) {
				registry[unknow.singleton] = { instance: type, resolve: null };
			} else {
				throw new Error("Type " + type.name + " doesn't exports any interface, please use registerAs")
			}
		}
	}

}
