import { Container, adapter, port, New } from './index';

// default use

function defaultUsage() {
    class Port {
        lower(text: string): string { throw 'Not implemented' }
    }

    @adapter(Port)
    class Adapter implements Port {
        lower(text: string): string {
            return text.toLowerCase();
        }
    }

    class Library {
        @port(Port)
        private port!: Port;

        test() {
            if (this.port.lower("UPPER") != "upper") {
                throw new Error("Test failed");
            }
            console.log("Test passed");
        }
    }

    Container.clear();
    Container.register(Adapter);
    let lib = new Library();
    lib.test();
}

function usageWithSymbol() {
    interface Port {
        lower(text: string): string;
    }

    const PortSymbol = Symbol("port");

    @adapter(PortSymbol)
    class Adapter implements Port {
        lower(text: string): string {
            return text.toLowerCase();
        }
    }

    class Library {
        @port(PortSymbol)
        private port!: Port;

        test() {
            if (this.port.lower("UPPER") != "upper") {
                throw new Error("Symbol test failed");
            }
            console.log("Symbol test passed");
        }
    }

    Container.clear();
    Container.register(Adapter);
    let lib = new Library();
    lib.test();
}

function singletons() {
    interface Port {
        count(): number;
    }

    const PortSymbol = Symbol("port");

    @adapter(PortSymbol, { singleton: true })
    class Adapter implements Port {
        number = 0;
        count(): number {
            return ++this.number;
        }

    }

    class Library {
        @port(PortSymbol)
        private port!: Port;

        test() {
            return this.port.count();
        }
    }

    Container.clear();
    Container.register(Adapter);
    const lib1 = new Library();
    lib1.test();
    const lib2 = new Library();
    if (lib1.test() != 2) {
        throw new Error("Singleton test1 failed");
    } else {
        console.log("Singleton test1 passed");
    }

    Container.clear();
    Container.registerSingleton(new Adapter(), PortSymbol);
    const lib3 = new Library();
    lib3.test();
    const lib4 = new Library();
    if (lib4.test() != 2) {
        throw new Error("Singleton test2 failed");
    } else {
        console.log("Singleton test2 passed");
    }
}

function usegeWithNameObjects() {
    const Port = {
        name: 'port'
    }

    @adapter(Port)
    class Adapter {
        lower(text: string): string {
            return text.toLowerCase();
        }
    }

    class Library {
        @port(Port)
        private port!: { lower: (string) => string };

        constructor() {
            if (this.port.lower("UPPER") != "upper") {
                throw new Error("Test failed");
            }
            console.log("Usage with names passed");
        }
    }

    Container.clear();
    Container.register(Adapter);
    let lib = new Library();
}

usegeWithNameObjects();
defaultUsage();
usageWithSymbol();
singletons();



function newOperator() {
    Container.clear();

    class Default {
        hello() { return "default" };
    }

    Container.singleton(Default);

    const def1 = New(Default);
    if (def1.hello() != 'default') throw new Error("Error in singleton with new");

    class Derive implements Default {
        hello() { return "derive" };
    }

    Container.clear();
    Container.singleton(Default, Derive);

    const def2 = New(Default);
    if (def2.hello() != 'derive') throw new Error("Error in singleton with new");

}

newOperator();